/*++

Copyright (c) 1995-98 Microsoft Corporation

Module Name:
    mqbench.cpp

Abstract:
    Benchmark MSMQ message performance

Author:
    Tatiana S.. (tatianas) 23-Feb-98

Environment:
	Platform-independent.

--*/

#include "mqbench.h"


/*
// Command line parsed parameters
*/
PARAMS g_Params;


/*
// Timer start interlocked, used to messure receive start time
// after waiting to first message
*/
int g_fTimerStarted = FALSE;

/*
// Synchronization Events. These events synchronize orchestrated start of all
// test threads.
*/
EVENTOBJ_OS g_hWaitToStart;
EVENTOBJ_OS g_hStart;
EVENTOBJ_OS g_hEnd;
EVENTOBJ_OS g_hSync1;
EVENTOBJ_OS g_hSync2;

/*
// Thread counter used to indentify the last thread in the bunch
*/
LONG g_ThreadCounter;

/*
// Benchmarck start and end time.
*/
TIMESTRUCT_OS g_StartTime;
TIMESTRUCT_OS g_EndTime;

void ErrorExit (char *pszErrorMsg, HRESULT hr)
{
	printf ("%s, %ul(%#x)\n", pszErrorMsg, hr,hr);
	exit(-1);
}

void InitParams(void)
{
	g_Params.Operation = SEND;
	g_Params.ulMsgNumber = 0;
	g_Params.ulTransactionNum = 1;
	g_Params.ulThreadNum = 1;
	g_Params.ulMsgSize = 0;
	g_Params.fWaitForReceivedMsg = FALSE;
	g_Params.fOperationSpecified = FALSE;
  g_Params.fClearQueue = FALSE;
  InitParams_I();
}

int IsValidParam (int Cur, int ParamNum, char *pszParams[])
{
	if (Cur+1==ParamNum)
		return FALSE;

	if (*pszParams[Cur+1] =='-')
		return FALSE;

	return TRUE;
}

void GetMsgType(char chMsgType)
{
	switch (chMsgType)
	{
	case 'e':
		g_Params.MsgType = EXPRESS;
		break;

	case 'r':
		g_Params.MsgType = RECOVERABLE;
		break;

	default:
		printf ("Invalid message type.\n");
		Usage_I();
	}
}

void GetMessageParams(int Cur, int ParamNum, char *pszParams[])
{
	/*
	// Get number of messages
	*/
	if (!IsValidParam (Cur, ParamNum, pszParams))
	{
		printf ("Missing number of messages.\n");
		Usage_I();
	}

	g_Params.ulMsgNumber = atol(pszParams[Cur+1]);
	if (g_Params.ulMsgNumber == 0) 
	{
		printf ("Invalid number of messages.\n");
		Usage_I();
	}
	
	/*
	// Get size of body
	*/
	if (!IsValidParam (Cur+1, ParamNum, pszParams))
	{
		printf ("Missing size of message body.\n");
		Usage_I();
	}
	g_Params.ulMsgSize = atol(pszParams[Cur+2]);
	if (g_Params.ulMsgNumber == 0) 
	{
		printf ("Invalid size of message body.\n");
		Usage_I();
	}
}

int ParseSendParams (int Cur, int ParamNum, char *pszParams[])
{
  char *pszMsgType;

	if (g_Params.fOperationSpecified)
	{
		printf ("Invalid operation. Valid value is either SEND or RECEIVE.\n");
		Usage_I();
	}
	
	g_Params.Operation = SEND;
	g_Params.fOperationSpecified = TRUE;
	
	/*
	// Get message type
	*/
	pszMsgType = pszParams[Cur];
	GetMsgType ((char)tolower(pszMsgType[1]));
	GetMessageParams (Cur, ParamNum, pszParams);
	
	return Cur + 3;
}

int ParseReceiveParams (int Cur, int ParamNum, char *pszParams[])
{
	if (g_Params.fOperationSpecified)
	{
		printf ("Invalid operation. Valid value is either SEND or RECEIVE.\n");
		Usage_I();
	}
	
	g_Params.Operation = RECEIVE;
	g_Params.fOperationSpecified = TRUE;
	
	GetMessageParams (Cur, ParamNum, pszParams);

	return Cur + 3;
}

int ParsePathParams(int Cur, int ParamNum, char *pszParams[])
{
	if (!IsValidParam(Cur, ParamNum, pszParams))
	{
		printf ("Missing queue path name.\n");
		Usage_I();
	}		

  ParsePathParams_I(pszParams[Cur+1],APP_Q);

	return Cur + 2;
}

int ParseClearParams(int Cur, int ParamNum, char *pszParams[])
{
	g_Params.fClearQueue = TRUE;
	return Cur + 1;
}

void GetSyncType(char chSyncType)
{
	switch (chSyncType)
	{
	case 'r':
		g_Params.SyncMode = SYNC_RECEIVE_WHILE_SENDING;
		break;

	case 's':
		g_Params.SyncMode = SYNC_SEND_THEN_RECEIVE;
		break;

	default:
		printf ("Invalid sync type.\n");
		Usage_I();
	}
}

void GetSyncParams(int Cur, int ParamNum, char *pszParams[]) {
	/*
	// Get sync queues
	*/
	if (!IsValidParam (Cur, ParamNum, pszParams))
	{
		printf ("Missing senderQ.\n");
		Usage_I();
	}

	if (!IsValidParam (Cur+1, ParamNum, pszParams))
	{
		printf ("Missing receiverQ.\n");
		Usage_I();
	}

  ParsePathParams_I(pszParams[Cur+1],SENDSYNC_Q);
  ParsePathParams_I(pszParams[Cur+2],RECVSYNC_Q);
}

int ParseSyncParams(int Cur, int ParamNum, char *pszParams[]) {
	/*
	// Get message type
	*/
	char *pszSyncType = pszParams[Cur];
	GetSyncType ((char)tolower(pszSyncType[1]));
	GetSyncParams (Cur, ParamNum, pszParams);

	return Cur + 3;
}

int ParseThreadParams(int Cur, int ParamNum, char *pszParams[])
{
	if (!IsValidParam(Cur, ParamNum, pszParams))
	{
		printf ("Missing number of threads.\n");
		Usage_I();
	}

	g_Params.ulThreadNum = atol(pszParams[Cur+1]);
	if (g_Params.ulThreadNum == 0)  
	{
		printf("Invalid number of threads.\n");
		Usage_I();
	}

	return Cur + 2;
}

int ParseWaitParams(int Cur)
{
	g_Params.fWaitForReceivedMsg = TRUE;
	return Cur + 1;
}

void ParseParams (int ParamNum, char *pszParams[])
{
  BOOL fDiagnostics = FALSE;
	int i = 1;

	if (ParamNum < 6)
	{
		Usage_I();
	}

	while (i < ParamNum)
	{
		if (*pszParams[i] != '-')
		{
			printf ("Invalid parameters\n");
			Usage_I();
		}

		switch (tolower(*(++pszParams[i])))
		{
		case 's':
			i = ParseSendParams(i, ParamNum, pszParams);
			break;

		case 'r':
			i = ParseReceiveParams(i, ParamNum, pszParams);
			break;
		
		case 'q':
			i = ParsePathParams(i, ParamNum, pszParams);
			break;

		case 'c':
			i = ParseClearParams(i, ParamNum, pszParams);
			break;

		case 'z':
			i = ParseSyncParams(i, ParamNum, pszParams);
			break;

		case 'x':
			i = ParseTransactionParams_I(i, ParamNum, pszParams);
			break;

		case 't':
      if(g_MultiThreadSupport) {
			  i = ParseThreadParams(i, ParamNum, pszParams);
      } else {
  			Usage_I();
      }
			break;
			
		case 'w':
			i = ParseWaitParams(i);
			break;

    case 'p':
      fDiagnostics = TRUE;
      i++;
      break;

		case '?':
			Usage_I();
			break;

		default:
			printf("Unknown switch %s\n", pszParams[i]);
			Usage_I();
			break;
		}
	}

	/*
	// check parameters
	*/
	if (!g_Params.fOperationSpecified)
	{
		printf("Send or Receive operation not speciifed\n");
		Usage_I();
	}

  if (!IsValidPathname_I())
	{
		printf("Queue path name not specified\n");
		Usage_I();
	}
	
  if(fDiagnostics) {
	  printf ("Operation:              %s\n", OPER2STR(g_Params.Operation));
    if(g_Params.Operation == SEND) {
	    printf ("Message type:           %s\n", MSGTYPE2STR(g_Params.MsgType));
    }
	  printf ("Number of messages:     %lu\n", g_Params.ulMsgNumber);
	  printf ("Size of messages:       %lu\n", g_Params.ulMsgSize);							
    PrintQueuename_I(APP_Q);
	  printf ("Transaction type:       %s\n", XACTTYPE2STR_I(g_Params.XactType));
	  printf ("Number of transactions: %d\n", g_Params.ulTransactionNum);
	  printf ("Number of threads:      %d\n", g_Params.ulThreadNum);
    printf ("Sync Mode:              %s\n", SYNCMODE2STR(g_Params.SyncMode)); 
    if(g_Params.SyncMode != NO_SYNC) {
      PrintQueuename_I(SENDSYNC_Q);
      PrintQueuename_I(RECVSYNC_Q);
    }
    printf ("ClearQueue:             %s\n", g_Params.fClearQueue ? "TRUE" : "FALSE"); 
  }
}

char* OPER2STR(DWORD dwOper) {
  char* psz;
  switch (dwOper) {
    DEF2STR(SEND);
    DEF2STR(RECEIVE);
    default: psz = "UNKNOWN_OPERATION";
  }
  return psz;
}

char* MSGTYPE2STR(DWORD dwMsgType) {
  char* psz;
  switch (dwMsgType) {
    DEF2STR(EXPRESS);
    DEF2STR(RECOVERABLE);
    default: psz = "UNKNOWN_MSGTYPE";
  }
  return psz;
}

char* SYNCMODE2STR(DWORD dwSyncMode) {
  char* psz;
  switch (dwSyncMode) {
    DEF2STR(SYNC_SEND_THEN_RECEIVE);
    DEF2STR(NO_SYNC);
    DEF2STR(SYNC_RECEIVE_WHILE_SENDING);
    default: psz = "UNKNOWN_MSGTYPE";
  }
  return psz;
}

void ResultOutput(void)
{
	char* pszOperation;
  ULONG nTotalMessages;
  double Time;
  double Benchmark;
  double Throughput;

	if (g_Params.Operation==SEND)
	{
		pszOperation = "Sent";
	}
	else
	{
		pszOperation = "Received";
	}

	nTotalMessages =	g_Params.ulMsgNumber * 
							      g_Params.ulTransactionNum * 
							      g_Params.ulThreadNum;

	printf("\nTotal messages:\t%lu %s\n", nTotalMessages, pszOperation);
	
	/*
	//convert to seconds
	*/
	
	Time = GetElapsedTime_OS(&g_StartTime,&g_EndTime);
	printf("Test time:\t%.3lf seconds\n", Time);

	Benchmark = nTotalMessages / Time;
	printf("Benchmark:\t%.0lf messages per second\n", Benchmark);

	Throughput = nTotalMessages * g_Params.ulMsgSize / Time;
	printf("Throughput:\t%.0lf bytes per second\n", Throughput);
}

DWORD
APIENTRY
TestThread(
	PVOID pv
	)
{
  HANDLE    hQueue;
  SYNCDATA  SyncData;
  BOOL      fSyncThread = FALSE;
  ULONG     ulTransCount;
  ULONG     ulMsgCount;

  hQueue = GetQueueHandle_I();

  /* set event that the queue has been opened and a connection
     is established with the queue manager */
	SetEvent_OS(g_hWaitToStart);

	if(InterlockedDecrement_OS(&g_ThreadCounter) == 0)
	{
    /* Check to see if the clear queue flag is enabled */
    if(g_Params.fClearQueue) {
      DWORD dwClearCount = ClearQueue_I(hQueue);

      if(dwClearCount) {
        printf("Note: %d messages were cleared from the benchmark queue.\n",dwClearCount);
      }
    }
    
    /* Check to see if Syncing is enabled */
    if(g_Params.SyncMode != NO_SYNC) {
      BeginSync(hQueue, &SyncData);

      /* mark this thread as the syncing thread */
      fSyncThread = TRUE;
    }

		/*
		// Last test thread sample start time, and enable all test
		// threads to run.
		*/
		GetTime_OS(&g_StartTime);
		g_ThreadCounter = g_Params.ulThreadNum;
    if(g_Params.fWaitForReceivedMsg) {
      printf("--Ready to begin timing--\n");
    }
		SetEvent_OS(g_hStart);
	}

	/*
	// Synchronize all threads to wait for a start signal
	*/
	WaitForEvent_OS(g_hStart);

	for (	ulTransCount = 0; 
			ulTransCount < g_Params.ulTransactionNum; 
			ulTransCount++ )
	{
		/*
		//BeginTransaction
		*/
		PTRANSOBJ pXact = GetTransactionPointer_I();

		/*
		// Send/Receive Messages
		*/
		for (	ulMsgCount=0; 
				ulMsgCount < g_Params.ulMsgNumber; 
				ulMsgCount++)
		{
			if (g_Params.Operation == SEND)
			{
        SendMsg_I(pXact,hQueue);
			}
			else
			{
        RecvMsg_I(pXact,hQueue);

				if(g_Params.fWaitForReceivedMsg && !g_fTimerStarted)
				{
					/*
					//	Start time is sampled after first receive, only by the
					//  first thread.
					*/
					if(InterlockedExchange_OS((LONG *)&g_fTimerStarted, TRUE) == FALSE)
					{
						GetTime_OS(&g_StartTime);
					}
				}
			}
		}
		
    CommitTrans_I(pXact,hQueue);
	}

	/*
	// Last thread sample end time and signal main thread to continue.
	*/
	if(InterlockedDecrement_OS(&g_ThreadCounter) == 0)
	{
		GetTime_OS(&g_EndTime);
		SetEvent_OS(g_hEnd);
	}


  /* terminate and cleanup sync data */
  if(fSyncThread) {
    /* Wait for main thread to signal to complete syncing */
    WaitForEvent_OS(g_hSync1);

    EndSync(hQueue, &SyncData);

    /* signal main thread that syncing is complete */
		SetEvent_OS(g_hSync2);
  }

  CloseQueue_I(&hQueue);

	return 0;
}

void RunTest(void)
{
	/*
	// Set thread counter to control TestThread.
	*/
	g_ThreadCounter = g_Params.ulThreadNum;

  StartThreads_OS();

	/*
	// Wait for all test threads to complete
	*/
	WaitForEvent_OS(g_hEnd);

  if(g_Params.SyncMode != NO_SYNC) {
    /* signal thread that is in sync mode to complete the sync operation */
		SetEvent_OS(g_hSync1);

    /* Wait for this thread to terminate */
	  WaitForEvent_OS(g_hSync2);
  }
}

void BeginSync(HANDLE hQueue, SYNCDATA* pSyncData) {
  OpenSyncQueues_I(hQueue, pSyncData);

  if(g_Params.Operation == SEND) {
    PerfomBeginSendSync(hQueue, pSyncData);
  } else {
    PerfomBeginRecvSync(hQueue, pSyncData);
  }
}

void EndSync(HANDLE hQueue, SYNCDATA* pSyncData) {
  if(g_Params.Operation == SEND) {
    PerfomEndSendSync(hQueue, pSyncData);
  } else {
    PerfomEndRecvSync(hQueue, pSyncData);
  }

  CloseSyncQueues_I(hQueue, pSyncData);
}

void PerfomBeginSendSync(HANDLE hQueue, SYNCDATA* pSyncData) {
  /* retrieve sync message from receiver */
  GetSyncMessage_I(hQueue,pSyncData,pSyncData->pReceiverId,NULL,BEGINSYNC);

  /* put sync message to receiver */
  PutSyncMessage_I(hQueue,pSyncData,pSyncData->pSenderId,pSyncData->pReceiverId,NORMALSYNC);

  /* retrieve sync message from receiver */
  GetSyncMessage_I(hQueue,pSyncData,NULL,pSyncData->pSenderId,NORMALSYNC);
  /* if the sync mode is set for receiver to be running, then sleep a */
  /* short while to let the receiver to notify all threads and to be ready */
  if(g_Params.SyncMode == SYNC_RECEIVE_WHILE_SENDING) {
    Sleep_OS(5000); /* sleep for 5 seconds */
  }
}

void PerfomBeginRecvSync(HANDLE hQueue, SYNCDATA* pSyncData) {
  /* put sync message to sender */
  PutSyncMessage_I(hQueue,pSyncData,pSyncData->pReceiverId,NULL,BEGINSYNC);

  /* retrieve sync message from sender */
  GetSyncMessage_I(hQueue,pSyncData,pSyncData->pSenderId,pSyncData->pReceiverId,NORMALSYNC);

  /* put sync message to sender */
  PutSyncMessage_I(hQueue,pSyncData,NULL,pSyncData->pSenderId,NORMALSYNC);

  /* if the sync mode is set for sender to complete first, then wait for the notify that
  /* the sender is complete */
  if(g_Params.SyncMode == SYNC_SEND_THEN_RECEIVE) {
    GetSyncMessage_I(hQueue,pSyncData,NULL,pSyncData->pReceiverId,NORMALSYNC);
  }
}

void PerfomEndSendSync(HANDLE hQueue, SYNCDATA* pSyncData) {
  /* put sync message to receiver to notify sender is finished if receiver */
  /* is waiting to start */
  if(g_Params.SyncMode == SYNC_SEND_THEN_RECEIVE) {
    PutSyncMessage_I(hQueue,pSyncData,NULL,pSyncData->pReceiverId,NORMALSYNC);
  }

  /* wait for receiver to finish */
  GetSyncMessage_I(hQueue,pSyncData,NULL,pSyncData->pSenderId,NORMALSYNC);
}

void PerfomEndRecvSync(HANDLE hQueue, SYNCDATA* pSyncData) {
  /* put sync message to sender to notify receiver is finished */
  PutSyncMessage_I(hQueue,pSyncData,NULL,pSyncData->pSenderId,NORMALSYNC);
}

void main(int argc, char *argv[])
{
	InitParams();
	
	ParseParams(argc, argv);

	InitTest_I();

	CreateEvents_OS();

	RunTest();

  CloseEvents_OS();

  CleanupTest_I();

	ResultOutput();

  exit(0);
}
