#define INITGUID

#include <string.h>

#ifdef FMQC_VER
#  include "fmqc.inl"
#else
#  include "msmq.inl"
#endif

#ifndef NO_EPAPI
#  include "cmqcv2.h"
#  include <MSMQExt.h>
#  include <MQSRExt.h>
#endif

/* global message defines */
#define x_PropCount 4

static MSGPROPID g_MessagePropId[x_PropCount] = {
	PROPID_M_BODY,
	PROPID_M_BODY_SIZE,
	PROPID_M_DELIVERY,
	PROPID_M_JOURNAL
};

static MQPROPVARIANT g_MessagePropVar[x_PropCount] = {
	{VT_VECTOR | VT_UI1, 0, 0, 0},
	{VT_UI4, 0, 0, 0},
	{VT_UI1, 0, 0, 0},
	{VT_UI1, 0, 0, 0}
};

static MQMSGPROPS g_MessageProperties = {
	x_PropCount,
	g_MessagePropId,
	g_MessagePropVar,
	0
};

void InitParams_I(void) {
	g_Params.MsgType = MQMSG_DELIVERY_EXPRESS;
	g_Params.XactType = NO_TRANSACTION;
	g_Params.I.wcsQueuePathName[0] = '\0';
}

int ParseTransactionParams_I(int Cur, int ParamNum, char *pszParams[])
{
	/*
	// Get transaction type
	*/
	char *pszXactType = pszParams[Cur];
	GetXactTypes((char)tolower(pszXactType[1]));
	
	/*
	// Get number of transaction
	*/ 
	if(g_Params.XactType == SINGLE)
	{
		return Cur + 1;
	}

	if (!IsValidParam (Cur, ParamNum, pszParams))
	{
		printf ("Missing number of transactions\n");
		Usage_I();
	}

	g_Params.ulTransactionNum = atol(pszParams[Cur+1]);
	if (g_Params.ulTransactionNum == 0)
	{
		printf("Invalid number of transaction.\n");
		Usage_I();
	}

	return Cur + 2;
}

char* XACTTYPE2STR_I(DWORD dwXactType) {
  char* psz;
  switch (dwXactType) {
    DEF2STR(NO_TRANSACTION);
    DEF2STR(INTERNAL);
    DEF2STR(COORDINATED);
    DEF2STR(SINGLE);
    default: psz = "UNKNOWN_XACTTYPE";
  }
  return psz;
}

void InitTest_I(void)
{
	SetMessageProps_I();

	if (g_Params.XactType == COORDINATED)
	{
		TransactionInit();
	}
}

void SetMessageProps_I(void)
{
	/*
	// Set the message body buffer
	*/
	assert(g_MessagePropId[0] == PROPID_M_BODY);

	g_MessagePropVar[0].MQVAL(caub.cElems) = g_Params.ulMsgSize;
	g_MessagePropVar[0].MQVAL(caub.pElems) = 
          (UCHAR*) malloc(sizeof(UCHAR)* g_Params.ulMsgSize);
	
	if (g_Params.Operation == SEND)
	{
		/*
		// Build message body
		*/
		memset(g_MessagePropVar[0].MQVAL(caub.pElems), 'a', g_Params.ulMsgSize);
	}

	/*
	// Set message body size
	*/
	assert(g_MessagePropId[1] == PROPID_M_BODY_SIZE);
	g_MessagePropVar[1].MQVAL(ulVal) = g_Params.ulMsgSize;

	/*
	// Set message delivery
	*/
	assert(g_MessagePropId[2] == PROPID_M_DELIVERY);
	g_MessagePropVar[2].MQVAL(bVal) = 
      g_Params.MsgType == EXPRESS ? MQMSG_DELIVERY_EXPRESS :
                                    MQMSG_DELIVERY_RECOVERABLE;

	/*
	// Set dead letter property
	*/
	assert(g_MessagePropId[3] == PROPID_M_JOURNAL);
	g_MessagePropVar[3].MQVAL(bVal) = MQMSG_DEADLETTER;
}

#define MAX_RETRIES 5
QUEUEHANDLE OpenQueue(WCHAR* wcsPathName, DWORD dwAccess) {
	HRESULT hr;
  QUEUEHANDLE hQueue;
	DWORD dwFormatNameLength = MAX_PATH;
	WCHAR wcsFormatName[MAX_PATH];
  int i;

  for(i=0;i<MAX_RETRIES;i++) {	
    if(i>0) {
      printf("retrying....\n");
      Sleep_OS(SLEEP_FACTOR);
    }
	  hr= MQPathNameToFormatName(
			  wcsPathName, 
			  wcsFormatName, 
			  &dwFormatNameLength
			  );
    if(hr != MQ_ERROR) break;
  }

	if (FAILED(hr)) 
	{
		ErrorExit("MQPathNameToFormatName failed", hr);
	}

	hr = MQOpenQueue(
			wcsFormatName,
			dwAccess,
			MQ_DENY_NONE,
			&hQueue
			);

	if (FAILED(hr)) 
	{
		ErrorExit("MQOpenQueue failed", hr);
	}

  return hQueue;
}


HANDLE GetQueueHandle_I(void)
{
	DWORD dwAccess;
  
	if (g_Params.Operation==SEND)
	{
		dwAccess = MQ_SEND_ACCESS;
	}
	else
	{
		dwAccess = MQ_RECEIVE_ACCESS;
	}
			
  return (HANDLE) OpenQueue(g_Params.I.wcsQueuePathName, dwAccess);
}


PTRANSOBJ GetTransactionPointer_I(void)
{
	ITransaction* pXact;
	HRESULT hr = MQ_OK;
	switch (g_Params.XactType)
	{
		case INTERNAL:
			hr = MQBeginTransaction (&pXact);
			break;

#ifndef FMQC_VER
		case COORDINATED:
			hr = DTCBeginTransaction(&pXact);
			break;
#endif

		case SINGLE:
			pXact = MQ_SINGLE_MESSAGE;
			break;

		case NO_TRANSACTION:
			pXact = MQ_NO_TRANSACTION;
			break;

		default:
			assert(0);
	}

	if (FAILED(hr)) 
	{
		ErrorExit("Can not create transaction", hr);
	}

	return (PTRANSOBJ)pXact;
}

void SendMsg_I(PTRANSOBJ pXact,HANDLE hQueue) {
	HRESULT hr;

	hr = MQSendMessage(	
			(QUEUEHANDLE)hQueue,
			&g_MessageProperties,
			(ITransaction*)pXact
			);

	if (FAILED(hr)) 
	{
   	CloseQueue_I(&hQueue);
		ErrorExit("MQSendMessage failed", hr);
	}
}

void RecvMsg_I(PTRANSOBJ pXact, HANDLE hQueue) {
  HRESULT hr;

	hr = MQReceiveMessage(
			(QUEUEHANDLE)hQueue,
			INFINITE,			/*dwTimeout*/
			MQ_ACTION_RECEIVE,  /* dwAction, */
			&g_MessageProperties,
			NULL,				/*IN OUT LPOVERLAPPED lpOverlapped, */
			NULL,				/*IN PMQRECEIVECALLBACK fnReceiveCallback, */
			NULL,				/*IN HANDLE hCursor, */
			(ITransaction*)pXact
			);

	if (FAILED(hr)) 
	{
   	CloseQueue_I(&hQueue);
		ErrorExit("MQReceiveMessage failed", hr);
	}
}

void CommitTrans_I(PTRANSOBJ pXact, HANDLE hQueue) {
	HRESULT hr;

	if (g_Params.XactType == INTERNAL || 
		g_Params.XactType == COORDINATED ) 
	{
    ITransaction* pX = (ITransaction*)pXact;
		hr = COMMIT(  pX,0,0,0);
		if (FAILED(hr))
		{
			ErrorExit("Commit failed", hr);
    	CloseQueue_I(&hQueue);
		}
		RELEASE(((ITransaction*)pXact));
	}
}

void CloseQueue_I(HANDLE* phQueue) {
  assert(phQueue != NULL);

 	MQCloseQueue((QUEUEHANDLE)*phQueue);

  *phQueue = NULL;
}

void GetXactTypes(char chXactType)
{
	switch (chXactType)
	{
	case 'i':
		g_Params.XactType = INTERNAL;
		break;

#ifndef FMQC_VER
	case 'c':
		g_Params.XactType = COORDINATED;
		break;
#endif

	case 's':
		g_Params.XactType = SINGLE;
		break;

	default:
		printf ("Invalid transaction type.\n");
		Usage_I();
	}
}

void PrintQueuename_I(QTYPE qType) {
  switch(qType) {
    case APP_Q:
    	_tprintf (_T("Queue Name:             %s\n"), g_Params.I.wcsQueuePathName);
      break;
   
    case SENDSYNC_Q:
    	_tprintf (_T("Send Sync Queue Name:   %s\n"), g_Params.I.wcsSendSyncQueue);
      break;

    case RECVSYNC_Q:
    	_tprintf (_T("Recv Sync Queue Name:   %s\n"), g_Params.I.wcsRecvSyncQueue);
      break;
  }
}

DWORD ClearQueue_I(HANDLE hQueue) {
  DWORD dwCount = 0;
  HRESULT hr = MQ_OK;
  ITransaction* pTrans = g_Params.XactType == NO_TRANSACTION ?  MQ_NO_TRANSACTION :
                                                                MQ_SINGLE_MESSAGE;
  QUEUEHANDLE hReceiveQ = OpenQueue(g_Params.I.wcsQueuePathName, MQ_RECEIVE_ACCESS);

  while(hr == MQ_OK) {
    hr = MQReceiveMessage(hReceiveQ,0,MQ_ACTION_RECEIVE,NULL,NULL,NULL,0,pTrans);
    if (hr == MQ_OK) {
      dwCount++;
    }
  }

  CloseQueue_I((HANDLE*)&hReceiveQ);

  return dwCount;
}

void OpenSyncQueues_I(HANDLE hQueue, SYNCDATA* pSyncData) {
  DWORD dwSyncSendAccess;
  DWORD dwSyncRecvAccess;

	if (g_Params.Operation==SEND)
	{
		dwSyncSendAccess = MQ_RECEIVE_ACCESS;
		dwSyncRecvAccess = MQ_SEND_ACCESS;
	}
	else
	{
		dwSyncSendAccess = MQ_SEND_ACCESS;
		dwSyncRecvAccess = MQ_RECEIVE_ACCESS;
	}
			
  pSyncData->hSendSyncQ = (HANDLE) OpenQueue( g_Params.I.wcsSendSyncQueue, 
                                              dwSyncSendAccess);
  pSyncData->hRecvSyncQ = (HANDLE) OpenQueue( g_Params.I.wcsRecvSyncQueue, 
                                              dwSyncRecvAccess);
}

void CloseSyncQueues_I(HANDLE hQueue, SYNCDATA* pSyncData) {
  CloseQueue_I(&(pSyncData->hSendSyncQ));
  CloseQueue_I(&(pSyncData->hRecvSyncQ));
}

void PrintMSGID(char* ptr) {
  int i;
  if(!ptr) {
    printf("NULL BUFFER");
    return;
  }
  for (i=0;i<20;i++) printf("%.2x ",ptr[i]);
}

void GetSyncMessage_I(HANDLE hQueue, SYNCDATA* pSyncData,char* pMsgId, char* pCorrId, 
                      SYNCOPERATION SyncOper) {
  HRESULT       hr;
  QUEUEHANDLE   hReceiveQ;
  enum {NumProps = 5};
  static MSGPROPID     aPropID[NumProps] = {PROPID_M_BODY,PROPID_M_EXTENSION,PROPID_M_EXTENSION_LEN,
                                       PROPID_M_MSGID,PROPID_M_CORRELATIONID};
  static MQPROPVARIANT aPropVar[NumProps] = {{VT_VECTOR|VT_UI1,0,0,0,0},
                                        {VT_VECTOR|VT_UI1,0,0,0,0},
                                        {VT_UI4,0,0,0,0},
                                        {VT_VECTOR|VT_UI1,0,0,0,0},
                                        {VT_VECTOR|VT_UI1,0,0,0,0}};
  MQMSGPROPS    MessageProperies = {NumProps,aPropID,aPropVar,NULL};
  char*         pInputMsgId[PROPID_M_MSGID_SIZE];
  char*         pInputCorrId[PROPID_M_MSGID_SIZE];
  ULONG         ulBodySize = SYNCOPERTEXTSIZE;
  char          pMsgBody[SYNCOPERTEXTSIZE];
  BOOL          fCorrectMsg = FALSE;
#ifndef NO_EPAPI
  MQMD          MqMd = {MQMD_DEFAULT};
  const DWORD   dwExtBufferSize = 1024;
  DWORD         dwExtSize;
  char          szExt[dwExtBufferSize];
  HANDLE        hExt;
  HANDLE        hCursor;
#endif

  /* determine which queue handle we will be receiving from */
	if (g_Params.Operation==SEND)
	{
		hReceiveQ = pSyncData->hSendSyncQ;
	}
	else
	{
		hReceiveQ = pSyncData->hRecvSyncQ;
	}

  /* initialize message properties */
  aPropVar[0].MQVAL(caub.pElems) = (UCHAR*)pMsgBody;
  aPropVar[0].MQVAL(caub.cElems) = ulBodySize;
#ifndef NO_EPAPI
  aPropVar[1].MQVAL(caub.pElems) = (UCHAR*)szExt;
  aPropVar[1].MQVAL(caub.cElems) = dwExtBufferSize;
#endif
  aPropVar[3].MQVAL(caub.pElems) = (UCHAR*)pInputMsgId;
  aPropVar[3].MQVAL(caub.cElems) = PROPID_M_MSGID_SIZE;
  aPropVar[4].MQVAL(caub.pElems) = (UCHAR*)pInputCorrId;
  aPropVar[4].MQVAL(caub.cElems) = PROPID_M_MSGID_SIZE;

  while(!fCorrectMsg) {
	  hr = MQReceiveMessage(
			  hReceiveQ,
			  SYNC_TIMEOUT,			/*dwTimeout */
			  MQ_ACTION_RECEIVE,  /* dwAction, */
			  &MessageProperies,
			  NULL,				/*IN OUT LPOVERLAPPED lpOverlapped, */
			  NULL,				/*IN PMQRECEIVECALLBACK fnReceiveCallback, */
			  NULL,				/*IN HANDLE hCursor, */
			  NULL
			  );

	  if (FAILED(hr)) 
	  {
      CloseSyncQueues_I(hQueue, pSyncData);
   	  CloseQueue_I(&hQueue);
		  ErrorExit("MQReceiveMessage failed", hr);
	  }

#ifndef NO_EPAPI
    hr = EPOpen(&hExt, szExt, aPropVar[2].MQVAL(ulVal));
    if (FAILED(hr)) {
      CloseSyncQueues_I(hQueue, pSyncData);
   	  CloseQueue_I(&hQueue);
		  ErrorExit("EPOpen failed", hr);
    }

    hCursor = NULL;
    while((hr = EPGet(hExt, EP_NEXT_FIELD, &hCursor, NULL, NULL, &dwExtSize)) == MQ_OK) {
      GUID FieldGUID;
      char* pField = (char*)malloc(dwExtSize);

      hr = EPGet(hExt, EP_CURRENT_FIELD, &hCursor, &FieldGUID, pField, &dwExtSize);
  	  if (FAILED(hr)) {
        EPClose(&hExt);
        CloseSyncQueues_I(hQueue, pSyncData);
   	    CloseQueue_I(&hQueue);
		    ErrorExit("EPGet failed", hr);
      }

      if(!memcmp(&FieldGUID, &sg_MSMQExtMQMD, sizeof(GUID))) {
        memcpy(&MqMd,pField,sizeof(MQMD));
        free(pField);
        break;
      }
      free(pField);
    }

    EPClose(&hExt);
#endif
  
    /* check to see if this is the correct message */
    if( !memcmp(SYNCOPERTEXT[SyncOper],pMsgBody,ulBodySize) && 
        (!pCorrId || 
          ((pCorrId && 
            (!memcmp(pCorrId,aPropVar[4].MQVAL(caub.pElems),PROPID_M_MSGID_SIZE) 
#ifndef NO_EPAPI
              || !memcmp(pCorrId,MqMd.CorrelId,sizeof(MQBYTE24))
#endif
            ))
        ))) {
      fCorrectMsg = TRUE;
    }
  } /* end while */

  /* copy the MsgId if it was needed */
  if(pMsgId) {
#ifndef NO_EPAPI
    MQBYTE24 pEmptyMsgId1;
    MQBYTE24 pEmptyMsgId2;
    memset(pEmptyMsgId1,' ',sizeof(MQBYTE24));
    memset(pEmptyMsgId2,'\0',sizeof(MQBYTE24));

    if(!memcmp(MqMd.MsgId,pEmptyMsgId1,sizeof(MQBYTE24)) ||
       !memcmp(MqMd.MsgId,pEmptyMsgId2,sizeof(MQBYTE24))) {
      memcpy(pMsgId,aPropVar[3].MQVAL(caub.pElems),PROPID_M_MSGID_SIZE);
    }
    else {
      memcpy(pMsgId,MqMd.MsgId,sizeof(MQBYTE24));
    }
#else
    memcpy(pMsgId,aPropVar[3].MQVAL(caub.pElems),PROPID_M_MSGID_SIZE);
#endif
  }
}

void PutSyncMessage_I(HANDLE hQueue, SYNCDATA* pSyncData,char* pMsgId, char* pCorrId, 
                      SYNCOPERATION SyncOper) {
  HRESULT       hr;
  QUEUEHANDLE   hSendQ;
  enum { BodyProp,ExtensionProp,ReceiveTimeProp,JournalProp,MsgIdProp,CorrIdProp,NumProps};
  static MSGPROPID     aPropID[NumProps] = {PROPID_M_BODY,PROPID_M_EXTENSION,PROPID_M_TIME_TO_BE_RECEIVED,PROPID_M_JOURNAL,PROPID_M_MSGID,
                                       PROPID_M_CORRELATIONID};
  static MQPROPVARIANT aPropVar[NumProps] = {{VT_VECTOR|VT_UI1,0,0,0,0},
                                      {VT_VECTOR|VT_UI1,0,0,0,0},
                                      {VT_UI4,0,0,0,0},
                                      {VT_UI1,0,0,0,0},
                                      {VT_VECTOR|VT_UI1,0,0,0,0},
                                      {VT_VECTOR|VT_UI1,0,0,0,0}};
  MQMSGPROPS    MessageProperies = {NumProps,aPropID,aPropVar,NULL};
  char*         pInputMsgId[PROPID_M_MSGID_SIZE];
#ifndef NO_EPAPI
  MQMD          MqMd = {MQMD_DEFAULT};
  const DWORD   dwExtBufferSize = 1024;
  DWORD         dwExtSize;
  char          szExt[dwExtBufferSize];
  HANDLE        hExt;
#endif

  /* determine which queue handle we will be sending to */
	if (g_Params.Operation==SEND)
	{
		hSendQ = pSyncData->hRecvSyncQ;
	}
	else
	{
		hSendQ = pSyncData->hSendSyncQ;
	}

  /* initialize message properties */
  aPropVar[BodyProp].MQVAL(caub.pElems) = (UCHAR*)SYNCOPERTEXT[SyncOper];
  aPropVar[BodyProp].MQVAL(caub.cElems) = SYNCOPERTEXTSIZE;

  aPropVar[ReceiveTimeProp].MQVAL(ulVal) = SYNC_MSG_LIFETIME;

  aPropVar[MsgIdProp].MQVAL(caub.pElems) = (UCHAR*)pInputMsgId;
  aPropVar[MsgIdProp].MQVAL(caub.cElems) = PROPID_M_MSGID_SIZE;
  aPropVar[JournalProp].MQVAL(bVal) = MQMSG_DEADLETTER;

  /* check to see if a CorrId is being used.  If not decrement number of properties */
  if(pCorrId) {
    aPropVar[CorrIdProp].MQVAL(caub.pElems) = (UCHAR*)pCorrId;
    aPropVar[CorrIdProp].MQVAL(caub.cElems) = PROPID_M_MSGID_SIZE;
#ifndef NO_EPAPI
    memcpy(MqMd.CorrelId,pCorrId,sizeof(MQBYTE24));
#endif
  }
  else {
    MessageProperies.cProp--;
  }

#ifndef NO_EPAPI
  /* Set the Extension Property */
  hr = EPOpen(&hExt, NULL, 0);
	if (FAILED(hr)) 
	{
    CloseSyncQueues_I(hQueue, pSyncData);
   	CloseQueue_I(&hQueue);
		ErrorExit("EPOpen failed", hr);
	}
  
  hr = EPAdd(hExt, &sg_MSMQExtMQMD, &MqMd, sizeof(MQMD), NULL);
	if (FAILED(hr)) 
	{
    EPClose(&hExt);
    CloseSyncQueues_I(hQueue, pSyncData);
   	CloseQueue_I(&hQueue);
		ErrorExit("EPAdd failed", hr);
	}

  dwExtSize = dwExtBufferSize;
  hr = EPGetBuffer(hExt, szExt, &dwExtSize);
  EPClose(&hExt);
	if (FAILED(hr)) 
	{
    CloseSyncQueues_I(hQueue, pSyncData);
   	CloseQueue_I(&hQueue);
		ErrorExit("EPGetBuffer failed", hr);
	}

  aPropVar[ExtensionProp].MQVAL(caub.cElems) = dwExtSize;
  aPropVar[ExtensionProp].MQVAL(caub.pElems) = (UCHAR*)szExt;
#endif

	hr = MQSendMessage(hSendQ, &MessageProperies,	NULL );

	if (FAILED(hr)) 
	{
    CloseSyncQueues_I(hQueue, pSyncData);
   	CloseQueue_I(&hQueue);
		ErrorExit("MQSendMessage failed", hr);
	}

  /* copy the MsgId if it was needed */
  if(pMsgId) {
    memcpy(pMsgId,(char*)aPropVar[MsgIdProp].MQVAL(caub.pElems),PROPID_M_MSGID_SIZE);
  }
}
